/* file dump/patch utility for MSDOS2 */
/* written in Aztec C version 3.20e */
/* by Jon Dart, Dept. of Anthropology, UCSD C-001, La Jolla, CA  92093 */

#include "ctype.h"
#include "stdio.h"
#include "fcntl.h"
#include "ascii.h"
#include "truth.h"
#include "fixpath.h"
#include "sgtty.h"

#define Version "1.6 (06-Sep-86)"
#define SECTSIZE 128  /* physical sector size */
#define LSECTSIZE (long) SECTSIZE /* long version of SECTSIZE */
#define PATHSIZE 65   /* size of a DOS path name plus 1 */
#define MAXSTR 81     /* max. console input line length + 1 */
#define exitch '\004'
#define ShowSector() printf("[%d/%d]",sector,maxsec)
#define offset(sector) (((long) sector) & 0xFFFF)*128L
#define skipsp(s,c) while (((c = *s) == SPACE) || (c==TAB)) s++
#define keyin(c)  c = getc(stdin)

extern char *malloc();
extern long lseek();

int fd = -1;
unsigned char *buffer;
unsigned int sector,nread,cmdnum,maxsec;
char *FileName;
unsigned char cmdlist[] = "ACDEFHOSWX+-?";
struct sgttyb conopts;


Menu()

/* show menu on screen */

{
    printf("\nAvailable commands:\n\n");

    printf("--- Dump commands ---       -------- Patch Commands --------\n");
    printf(" A = Dump in ASCII           C = Copy Sector(s) to New File\n");
    printf(" D = Dump in Decimal         E = Edit Sector Buffer\n");
    printf(" H = Dump in Hex             W = Write Buffer to Disk\n");
    printf(" O = Dump in Octal\n");
    printf("\n");
    printf("-- Movement commands --     --- Misc. Commands ---\n");
    printf(" + = Next Sector             F = Select New File\n");
    printf(" - = Previous Sector         ? = Display Commands\n");
    printf(" S = Select Sector           X = Exit to DOS\n");
}

WriteShortHeader()

{
    printf("    0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F\n");
    printf("   ------------------------------------------------\n");
}

WriteLongHeader()

{
    printf("    0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F\n");
    printf("  ----------------------------------------------------------------\n");
}

ViewByte(mode,b)
char mode; unsigned char b;
{
    if (mode == 'A') { /* ascii mode */
        b = toascii(b);
        if (isprint(b))
            printf(" %c",b);
        else if (b!='\127') /* non-printing char */
            printf("^%c",b+'A'-1);
        else /* DEL char. */
            printf("^?");
    }
    else if (mode == 'D')
        printf("%3d",b);
    else if (mode == 'H')
        printf("%02x",b);
    else if (mode == 'O')
        printf("%03o",b);
}

ViewSector(mode)
char mode;

{
    unsigned int i,j,count;
    unsigned char *a;

    if (fd != -1) printf("%s ",FileName);
    ShowSector(); putchar(NL);
    if ((mode=='A') || (mode=='H'))
        WriteShortHeader();
    else
        WriteLongHeader();
    count = 0;
    for (i=0;i<8;i++) {
       printf("%1x |",i);
       for (j=0;j<16;j++) {
           if (++count<=nread) {
               ViewByte(mode,buffer[count]);
               putchar(' ');
           }
       }
       putchar(NL);
    }
}

GetLine(s,max,exitchar)
char *s; unsigned int max; char exitchar;

/* get line from keyboard. max = max # of chars, exitch = char. to cause
   immediate exit.  Returns 0 if exitch typed, 1 if not */

{
    unsigned int k,exit;
    char ch;

    exit = FALSE;
    k = 0;
    do {
        keyin(ch);
        if (ch == exitchar) {
            ViewByte('A',ch);
            exit = TRUE;
        }
        else if ((ch==BKS) || (ch==DEL)) {
            if (k>0) {
                --k;
                putchar(BKS);
                putchar(SPACE);
                putchar(BKS);
            }
        }
        else if (isprint(ch)) {
            if (k<max) {
                putchar(ch);
                s[k] = ch;
                ++k;
            }
        }
   }
   while ((ch!=NL) && (!exit));
   putchar(NL);
   s[k] = EOS;
   return(!exit);
}

GetFile(prompt,filename,mode)
char *prompt; char *filename; int mode;

/* get file name from console, open it (in mode) for unbuffered i/o, and
   return filename and file descriptor (as function value) */
/* returns -1 if no file opened */

{
    int gotfile, type;
    char sp[PATHSIZE],lip[PATHSIZE];
    char c;
    long int filelen;
    int fd;

    gotfile = FALSE;
    fd = -1;
    do {
        printf("%s",prompt);
        if (!GetLine(filename,PATHSIZE,exitch)) return;
        skipsp(filename,c);
        if (c!=EOS) {
            fd = open(filename,mode);
            if (fd == -1) printf("\nCan't open: %s\n",filename);
        }
   } while ((filename[0]!=EOS) && (fd == -1));
   return(fd);
}

NewFile()

/* open new file for viewing/patching */
/* returns its file descriptor (-1 if no file opened) */

{
    long filelen;
    int fd;

    fd = GetFile("File name? ",FileName,O_RDWR);
    if (fd != -1) {
        nread = read(fd,buffer,SECTSIZE);
        filelen = lseek(fd,0L,2);
        maxsec = (filelen/((long) SECTSIZE)) & 0xFFFF;
        lseek(fd,0L,0);
    }
    sector = 0;
    return(fd);
}

CmdSrc(s,ch)
register char *s; char ch;

/* look for ch in s, return index to 1st match (1..length(s)) or 0 if
   not found */

{
    int j;
    j = 1;
    while (*s) {
        if (*s==ch)
            return(j);
        else {
            j++;
            s++;
        }
    }
    return(0);
}

StringToNum(d,base,num)
char *d; unsigned int base; int *num;

/* converts number in string d (of given base) to integer num */
/* returns 0 if error */

{
     int n;

     switch(base) {
     case 10: n = sscanf(d,"%d",num); break;
     case 16: n = sscanf(d,"%x",num); break;
     case 8:  n = sscanf(d,"%o",num); break;
     default:
         return(0);
     }
     if (n==1)
         return(1);
     else
         return(0);
}

stuff(s,mode,base,buffer,i)
char *s; char mode; int base; unsigned char *buffer; int i;

/* read bytes from s, stuff into buffer at position i, return new value of i */

{
    int j,k,done,num;
    char d[MAXSTR];
    register char ch;

    j = 0; done = TRUE;
    if (mode=='A') { /* ascii mode */
        k = 0;
        while (s[j]) {
            k = (i+j) & (SECTSIZE-1);
            buffer[k] = s[j++];
        }
        i = (k+1) & (SECTSIZE-1);
    }
    else
    do {
        skipsp(s,ch);
        k = 0;
        /* see if data string ok */
        while ((k<MAXSTR) 
            && (((mode=='D') && isdigit(ch))
            || ((mode=='O') && (ch>='0') && (ch<='7'))
            || ((mode=='H') && (isdigit(ch) || ((ch>='A') && (ch<='F')))))) {
            d[k++] = ch;
            ch = *(++s);
        }
        if ((!isspace(ch)) && (ch!=EOS)) { /* non-numeric char. */
            printf("Bad character in input: %c\n",ch);
            return(i);
        }
        d[k]=EOS;
        if (k==0) /* nothing typed, just advance to next byte */
            i = (i+1) & (SECTSIZE-1);
        else { /* got a number in d */
            if (StringToNum(d,base,&num)) {
                buffer[i] = num;
                i = (i + 1) & (SECTSIZE-1);
            }
            else { /* conversion error */
                printf("Error in value: %s\n",d);
                done = FALSE;
            }
        }
    } while ((done) && (ch!=EOS));
    return(i);
}

EditSector(mode,start)
char mode; unsigned int start;

/* accept patch info from user and modify the buffer.  'mode' is
   'D', 'H' or 'O' (for decimal, hex, octal).  'start' is the
   index into the buffer for the first byte to be patched. */

{

    unsigned int base,i,j,k,num,exit;
    char s[MAXSTR],c,ch;
    unsigned char *a;

    switch(mode) {
       case 'D': base = 10; break;
       case 'H': base = 16; break;
       case 'O': base = 8;  break;
       default:
       /* base undefined */ ;
    }
    i = start; exit = FALSE;
    do {
        printf("%02x: ",i);
        a = buffer + i;
        ViewByte(mode,*a); printf(" -> ");
        if ((exit = !GetLine(s,MAXSTR,exitch))==FALSE)
            i = stuff(s,mode,base,buffer,i);
    }
    while (exit==FALSE);
}

MoveToSec(newsec)
unsigned int newsec;

/* seeks to a new sector, reads it into buffer, and sets global vars.
   appropriately.  Returns 0 if error */

{
    if (newsec>maxsec) newsec=maxsec;
    lseek(fd,offset(newsec),0);  /* rewind to start of file */
    nread = read(fd,buffer,SECTSIZE);
    lseek(fd,offset(newsec),0);
    if (nread == -1) {
        printf("\nDisk read error.\n");
        MoveToSec(sector); /* attempt to return to old sector */
        return(0);
    }
    else {
        sector=newsec;
        return(1);
    }
}

GetSecRange(sec1,sec2)
register int *sec1,*sec2;

/* get a single sector number or a range of sector numbers from the
   console. returns 0 if error */

{
    char d[20],sectnums[20];
    char *ds,*s;
    char c;
    int err;

    do {
        err = FALSE;
        printf("Copy which sectors (0-%d)? ",maxsec);
        if (!GetLine(sectnums,20,exitch))
            return(0);
        s = sectnums; ds = d;
        skipsp(s,c);
        if (c==EOS)
            return(0);
        while (isdigit(*s)) *ds++ = *s++;
        *ds = EOS;
        err = (sscanf(d,"%d",sec1) == 1) ? FALSE : TRUE;
        if (!err && (*s==EOS)) { /* only 1 num specified */
            *sec2 = *sec1;
        }
        if (!err && (*s=='-')) {
            ds = d;
            while (isdigit(*(++s))) *ds++ = *s;
            *ds = EOS;
            if (sscanf(d,"%d",sec2)!=1)
                err = TRUE;
        }
        if (err || (*sec1<0) || (*sec1>maxsec) || (*sec2<0) || (*sec2>maxsec)) {
           printf("\nEnter a single sector or a range of sectors (0-%d).\n",maxsec);
        }
    } while (err);
    return(!err);
}

CopySectors()

/* copy sector(s) from currently open file to another file */

{
    int fd2,sec1,sec2,nr,nw,i;
    char fn2[PATHSIZE];
    char c;
    char sectnums[20];
    char secbuf[SECTSIZE];
    char d[5];
    char *s,*ds;

    if (fd == -1) {
        printf("\nNo file open.\n");
        return;
    }
    if ((fd2 = GetFile("File name to copy to? ",fn2,O_CREAT+O_WRONLY)) == -1)
        return;
    if (!GetSecRange(&sec1,&sec2)) {
        close(fd2);
        return;
    }
    lseek(fd,offset(sec1),0);
    for (i=sec1;i<=sec2;i++) {
        nr = read(fd,secbuf,SECTSIZE);
        nw = write(fd2,secbuf,nr);
    }
    close(fd2);
    MoveToSec(sector);
}

Change()

/* change contents of sector buffer */

{
    char s[5];
    unsigned int num,base,done,exit,ok;
    static char options[] = "ADHO";
    char mode;

    printf("Format (A = Ascii, D = Decimal, H = Hex, O = Octal)? ");
    do {
        keyin(mode);
        mode = toupper(mode);
    }
    while ((CmdSrc(options,mode) == 0) && (mode!=NL));
    putchar(mode);
    if (mode != NL) {
        ok = FALSE;
        do {
            printf("\nStarting address within buffer (00-7F)? ");
            if (!GetLine(s,MAXSTR,exitch)) return;
            if (StringToNum(s,16,&num)) {
                ok = TRUE;
                if (mode=='A') {
                    printf("Enter text to replace existing sector contents.\n");
                    printf("Type ^D when done.\n");
                }
                else {
                    printf("Enter new bytes, one or more per line.\n");
                    printf("Use spaces to separate. Type ^D when done.\n");
                }
                EditSector(mode,num);
            }
        } while (!ok);
    }
}

Forward()

{
    if (sector+1<=maxsec) {
        if (MoveToSec(sector+1)) {
            ShowSector();
            putchar(NL);
        }
    }
    else
        printf("[END OF FILE]");
}

Reverse()

{
    if (sector > 0) {
       if (MoveToSec(sector-1)) {
       /* read ok */
           ShowSector(); putchar(NL);
       }
    }
}

SetSector()

{
    int newsec;
    char s[MAXSTR];

    if (maxsec>0) {
        printf("Sector (0-%d)? ",maxsec);
        if (GetLine(s,MAXSTR,exitch)) {
            if (StringToNum(s,10,&newsec)) {
                if (MoveToSec(newsec)) {
                    ShowSector();
                    putchar(NL);
                }
            }
            else
                printf("\nError - bad number\n");
        }
    }
}

Update()

{
    int nwritten;

    lseek(fd,offset(sector),0);
    nwritten = write(fd,buffer,SECTSIZE);
    lseek(fd,offset(sector),0);
    if (nwritten != SECTSIZE) {
       printf("\nError (probably disk full)\n");
    }
}

DoCommand(cmd)
char cmd;
{
    putchar(cmd);
    putchar(NL);
    switch(cmd) {
    case 'E': Change();    break;
    case 'C': CopySectors(); break;
    case 'A':
    case 'D':
    case 'H':
    case 'O': ViewSector(cmd); break;
    case 'S': SetSector();      break;
    case 'F': {
               if (fd != -1) close(fd);
               fd = NewFile(); break;
              }
    case 'W': Update();    break;
    case '+': Forward();   break;
    case '-': Reverse();   break;
    case '?': Menu();      break;
    default:
          /* nothing */ ;
    }
}

main(argc,argv)
int argc; char *argv[];
{
    struct regstruc {
       unsigned int ax,bx,cx,dx,si,di;
    } inregs, outregs;
    unsigned int ctlcstat;
    int gotfile;
    long filelen;
    char cmd;

    printf("File Dump/Patch Utility\n");
    printf("by Jon Dart ...  Version %s\n\n",Version);

    /* turn on char. mode, no echo */
#ifndef DEBUG
    ioctl(fileno(stdin),TIOCGETP,&conopts);
    conopts.sg_flags |= (CBREAK+CRMOD);
    conopts.sg_flags &= ~(ECHO);
    ioctl(fileno(stdin),TIOCSETP,&conopts);
#endif
    inregs.ax = 0;
    doscall(0x33,&inregs,&outregs);  /* get current ^C status */
    ctlcstat = outregs.dx; /* save it */
    inregs.ax = 1;
    inregs.dx = 0;
    doscall(0x33,&inregs,&outregs); /* turn off ^C trapping */

    buffer = (unsigned char *)  malloc(SECTSIZE);
    FileName = malloc(PATHSIZE);
    gotfile = FALSE;

    if (argc>1) {
        fd = open(argv[1],O_RDWR);
        gotfile = (fd != -1);
        if (!gotfile) {
            printf("\nError - can't open: %s\n",argv[1]);
        }
        else {
            nread = read(fd,buffer,SECTSIZE);
            filelen = lseek(fd,0L,2);
            maxsec = (filelen/((long) SECTSIZE)) & 0xFFFF;
            lseek(fd,0L,0);
            sector = 0;
            strcpy(FileName,argv[1]);
            printf("File %s\n",FileName);
        }
    }
    if (gotfile || (fd = NewFile()) != -1) { /* file open ok */
        Menu(); printf("\n->");
        do {
            clrco(); /* clear type-ahead */
            keyin(cmd);
            cmd = toupper(cmd);
            cmdnum = CmdSrc(cmdlist,cmd);
            if (cmdnum) { /* valid command, execute it */
                DoCommand(cmd);
                if (cmd != 'X') {
                    printf("\n->");
                }
            }
        }
        while (cmd!='X');
    }
    if (fd != -1) close(fd);
    free(buffer);

    /* restore console to line mode, with echo: */

    ioctl(fileno(stdin),TIOCGETP,&conopts);
    conopts.sg_flags &= ~(CBREAK+CRMOD);
    conopts.sg_flags |= ECHO;
    ioctl(fileno(stdin),TIOCSETP,&conopts);

    /* restore old ^C status */

    inregs.ax = 1;
    inregs.dx = ctlcstat;
    doscall(0x33,&inregs,&outregs);
}

